/*
Very simple filesystem support.

Directory listing:

<entry>
...
<entry with length set to 0>

Directory entry:

<length> <location> <metadata> <name string>

Obviously, the structures are not particularly amenable to modification, and
so the filesystem is mostly intended for read-only use.
*/

#include <string.h>

#include <fsclient/client.h>
#include <fsclient/fsops.h>
#include "fslib.h"



/* Type operations. */

static inline length_t get_length(char *data)
{
  return ((entry_t *) data)->length;
}

static inline location_t get_location(char *data)
{
  return ((entry_t *) data)->location;
}

static inline metadata_t get_metadata(char *data)
{
  return ((entry_t *) data)->metadata;
}

static inline char *get_name(char *data)
{
  return ((entry_t *) data)->n.namearray;
}

static inline void set_length(char *data, length_t length)
{
  ((entry_t *) data)->length = length;
}

static inline void set_location(char *data, location_t location)
{
  ((entry_t *) data)->location = location;
}

static inline void set_metadata(char *data, metadata_t metadata)
{
  ((entry_t *) data)->metadata = metadata;
}

static inline void set_name(char *data, const char *name)
{
  strcpy(get_name(data), name);
}



/* Storage navigation. */

/* Return whether the needed space is available in the region. */

static int have_needed(file_descriptor_t *region, location_t needed)
{
  return needed < fs_get_available(region);
}

/* Ensure that the current entry is in the region. */

static void ensure_entry(file_descriptor_t *region)
{
  location_t length;

  if (!have_needed(region, sizeof(length_t)))
    client_ensure_read_position(region, fs_get_data_start_position(region));

  length = get_length(fs_get_data_start(region));

  if (!have_needed(region, length))
    client_ensure_read_position(region, fs_get_data_start_position(region));
}

/* Ensure that space for a final entry is available in the region. */

static void ensure_final_entry(file_descriptor_t *region)
{
  if (!have_needed(region, sizeof(length_t)))
    client_ensure_read_position(region, fs_get_data_start_position(region));
}



/* Storage access. */

/* Obtain an entry, referencing the entry name, not copying it. */

void get_entry(char *data, entry_t *entry)
{
  entry->length = get_length(data);
  entry->location = get_location(data);
  entry->metadata = get_metadata(data);
  entry->n.name = get_name(data);
}

/* Advance the data pointer to the next entry. */

static void _next_entry(file_descriptor_t *region)
{
  region->data_start += get_length(fs_get_data_start(region));
}

/* Advance to the next entry, ensuring it is available. */

void next_entry(file_descriptor_t *region)
{
  _next_entry(region);
  ensure_entry(region);
}

/* Advance to what will be the final (null) entry, ensuring it can be set. */

void final_entry(file_descriptor_t *region)
{
  _next_entry(region);
  ensure_final_entry(region);
}

/* Find the named entry, populating the supplied entry structure, returning
   whether such an entry was found. */

int find_entry(file_descriptor_t *region, const char *name, entry_t *entry)
{
  length_t length;
  char *p = strstr(name, "/");
  size_t n = p != NULL ? p - name : strlen(name);
  char *found;

  while (1)
  {
    if (!(length = get_length(fs_get_data_start(region))))
      return 0;

    found = get_name(fs_get_data_start(region));

    if (!strncmp(found, name, n) && (strlen(found) == n))
    {
      get_entry(fs_get_data_start(region), entry);
      return 1;
    }

    region->data_start += length;
    ensure_entry(region);
  }

  return 0;
}

/* Find the named entry by recursing in the directory structure, taking each
   component of the name as a directory with the last component identifying any
   kind of object, returning whether an entry was found. */

int find_entry_recursively(file_descriptor_t *region, const char *name, entry_t *entry)
{
  char *p = strstr(name, "/");

  client_ensure_read_position(region, fs_get_data_start_position(region));


  if (find_entry(region, name, entry))
  {
    if (p == NULL)
    {
      return 1;
    }
    else
      if (entry->metadata.type == OBJTYPE_DIR)
      {
        client_ensure_read_position(region, entry->location);
        ensure_entry(region);
        return find_entry_recursively(region, p + 1, entry);
      }
  }

  return 0;
}

/* Seek to an entry having the given index, returning whether such an entry was
   present. */

int seek_entry(file_descriptor_t *region, int index)
{
  length_t length;
  int i = 0;

  while (1)
  {
    if (!(length = get_length(fs_get_data_start(region))))
      return 0;

    if (i == index)
      return 1;

    region->data_start += length;
    ensure_entry(region);
  }

  return 0;
}



/* Storage modification. */

/* Return the length of an entry having the given name. */

length_t get_entry_length(const char *name)
{
  return sizeof(entry_t) - sizeof(entry_name_t) + strlen(name) + 1;
}

/* Establish a new directory by recording the final (null) entry. */

void new_directory(file_descriptor_t *region)
{
  set_length(fs_get_data_start(region), 0);
}
